"""
Detection evaluation.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import itertools
import logging

import numpy as np

import fiftyone.core.evaluation as foe
import fiftyone.core.fields as fof
import fiftyone.core.labels as fol
import fiftyone.core.validation as fov

from .base import BaseEvaluationResults


logger = logging.getLogger(__name__)


def evaluate_detections(
    samples,
    pred_field,
    gt_field="ground_truth",
    eval_key=None,
    classes=None,
    missing=None,
    method=None,
    iou=0.50,
    use_masks=False,
    use_boxes=False,
    classwise=True,
    dynamic=True,
    **kwargs,
):
    """Evaluates the predicted detections in the given samples with respect to
    the specified ground truth detections.

    This method supports evaluating the following spatial data types:

    -   Object detections in :class:`fiftyone.core.labels.Detections` format
    -   Instance segmentations in :class:`fiftyone.core.labels.Detections`
        format with their ``mask`` attributes populated
    -   Polygons in :class:`fiftyone.core.labels.Polylines` format
    -   Temporal detections in :class:`fiftyone.core.labels.TemporalDetections`
        format

    For spatial object detection evaluation, this method uses COCO-style
    evaluation by default.

    For temporal segment detection, this method uses ActivityNet-style
    evaluation by default.

    You can use the ``method`` parameter to select a different method, and you
    can optionally customize the method by passing additional parameters for
    the method's config class as ``kwargs``.

    The supported ``method`` values and their associated configs are:

    -   ``"coco"``: :class:`fiftyone.utils.eval.coco.COCOEvaluationConfig`
    -   ``"open-images"``: :class:`fiftyone.utils.eval.openimages.OpenImagesEvaluationConfig`
    -   ``"activitynet"``: :class:`fiftyone.utils.eval.activitynet.ActivityNetEvaluationConfig`

    If an ``eval_key`` is provided, a number of fields are populated at the
    object- and sample-level recording the results of the evaluation:

    -   True positive (TP), false positive (FP), and false negative (FN) counts
        for each sample are saved in top-level fields of each sample::

            TP: sample.<eval_key>_tp
            FP: sample.<eval_key>_fp
            FN: sample.<eval_key>_fn

        In addition, when evaluating frame-level objects, TP/FP/FN counts are
        recorded for each frame::

            TP: frame.<eval_key>_tp
            FP: frame.<eval_key>_fp
            FN: frame.<eval_key>_fn

    -   The fields listed below are populated on each individual object; these
        fields tabulate the TP/FP/FN status of the object, the ID of the
        matching object (if any), and the matching IoU::

            TP/FP/FN: object.<eval_key>
                  ID: object.<eval_key>_id
                 IoU: object.<eval_key>_iou

    Args:
        samples: a :class:`fiftyone.core.collections.SampleCollection`
        pred_field: the name of the field containing the predicted
            :class:`fiftyone.core.labels.Detections`,
            :class:`fiftyone.core.labels.Polylines`,
            or :class:`fiftyone.core.labels.TemporalDetections`
        gt_field ("ground_truth"): the name of the field containing the ground
            truth :class:`fiftyone.core.labels.Detections`,
            :class:`fiftyone.core.labels.Polylines`,
            or :class:`fiftyone.core.labels.TemporalDetections`
        eval_key (None): an evaluation key to use to refer to this evaluation
        classes (None): the list of possible classes. If not provided, the
            observed ground truth/predicted labels are used
        missing (None): a missing label string. Any unmatched objects are given
            this label for results purposes
        method (None): a string specifying the evaluation method to use.
            For spatial object detection, the supported values are
            ``("coco", "open-images")`` and the default is ``"coco"``. For
            temporal segment detection, the supported values are
            ``("activitynet")`` and the default is ``"activitynet"``
        iou (0.50): the IoU threshold to use to determine matches
        use_masks (False): whether to compute IoUs using the instances masks in
            the ``mask`` attribute of the provided objects, which must be
            :class:`fiftyone.core.labels.Detection` instances
        use_boxes (False): whether to compute IoUs using the bounding boxes
            of the provided :class:`fiftyone.core.labels.Polyline` instances
            rather than using their actual geometries
        classwise (True): whether to only match objects with the same class
            label (True) or allow matches between classes (False)
        dynamic (True): whether to declare the dynamic object-level attributes
            that are populated on the dataset's schema
        **kwargs: optional keyword arguments for the constructor of the
            :class:`DetectionEvaluationConfig` being used

    Returns:
        a :class:`DetectionResults`
    """
    fov.validate_collection_label_fields(
        samples,
        (pred_field, gt_field),
        (fol.Detections, fol.Polylines, fol.TemporalDetections),
        same_type=True,
    )

    label_type = samples._get_label_field_type(gt_field)
    is_temporal = issubclass(label_type, fol.TemporalDetections)

    if is_temporal:
        fov.validate_video_collection(samples)
    else:
        kwargs.update(dict(use_masks=use_masks, use_boxes=use_boxes))

    config = _parse_config(
        pred_field,
        gt_field,
        method,
        is_temporal,
        iou=iou,
        classwise=classwise,
        **kwargs,
    )

    eval_method = config.build()
    eval_method.ensure_requirements()

    eval_method.register_run(samples, eval_key)
    eval_method.register_samples(samples, eval_key, dynamic=dynamic)

    processing_frames = samples._is_frame_field(pred_field)

    if eval_key is not None:
        tp_field = "%s_tp" % eval_key
        fp_field = "%s_fp" % eval_key
        fn_field = "%s_fn" % eval_key

    if config.requires_additional_fields:
        _samples = samples
    else:
        _samples = samples.select_fields([gt_field, pred_field])

    matches = []
    logger.info("Evaluating detections...")
    for sample in _samples.iter_samples(progress=True):
        if processing_frames:
            docs = sample.frames.values()
        else:
            docs = [sample]

        sample_tp = 0
        sample_fp = 0
        sample_fn = 0
        for doc in docs:
            doc_matches = eval_method.evaluate(doc, eval_key=eval_key)
            matches.extend(doc_matches)
            tp, fp, fn = _tally_matches(doc_matches)
            sample_tp += tp
            sample_fp += fp
            sample_fn += fn

            if processing_frames and eval_key is not None:
                doc[tp_field] = tp
                doc[fp_field] = fp
                doc[fn_field] = fn

        if eval_key is not None:
            sample[tp_field] = sample_tp
            sample[fp_field] = sample_fp
            sample[fn_field] = sample_fn
            sample.save()

    results = eval_method.generate_results(
        samples, matches, eval_key=eval_key, classes=classes, missing=missing
    )
    eval_method.save_run_results(samples, eval_key, results)

    return results


class DetectionEvaluationConfig(foe.EvaluationMethodConfig):
    """Base class for configuring :class:`DetectionEvaluation` instances.

    Args:
        pred_field: the name of the field containing the predicted
            :class:`fiftyone.core.labels.Detections` or
            :class:`fiftyone.core.labels.Polylines`
        gt_field: the name of the field containing the ground truth
            :class:`fiftyone.core.labels.Detections` or
            :class:`fiftyone.core.labels.Polylines`
        iou (None): the IoU threshold to use to determine matches
        classwise (None): whether to only match objects with the same class
            label (True) or allow matches between classes (False)
    """

    def __init__(
        self, pred_field, gt_field, iou=None, classwise=None, **kwargs
    ):
        super().__init__(**kwargs)
        self.pred_field = pred_field
        self.gt_field = gt_field
        self.iou = iou
        self.classwise = classwise

    @property
    def requires_additional_fields(self):
        """Whether fields besides ``pred_field`` and ``gt_field`` are required
        in order to perform evaluation.

        If True then the entire samples will be loaded rather than using
        :meth:`select_fields() <fiftyone.core.collections.SampleCollection.select_fields>`
        to optimize.
        """
        return False


class DetectionEvaluation(foe.EvaluationMethod):
    """Base class for detection evaluation methods.

    Args:
        config: a :class:`DetectionEvaluationConfig`
    """

    def __init__(self, config):
        super().__init__(config)
        self.gt_field = None
        self.pred_field = None

    def register_samples(self, samples, eval_key, dynamic=True):
        """Registers the collection on which evaluation will be performed.

        This method will be called before the first call to :meth:`evaluate`.
        Subclasses can extend this method to perform any setup required for an
        evaluation run.

        Args:
            samples: a :class:`fiftyone.core.collections.SampleCollection`
            eval_key: the evaluation key for this evaluation
            dynamic (True): whether to declare the dynamic object-level
                attributes that are populated on the dataset's schema
        """
        _gt_field = self.config.gt_field
        _pred_field = self.config.pred_field

        self.gt_field, _ = samples._handle_frame_field(_gt_field)
        self.pred_field, _ = samples._handle_frame_field(_pred_field)

        if eval_key is None:
            return

        processing_frames = samples._is_frame_field(_pred_field)
        dataset = samples._dataset

        tp_field = "%s_tp" % eval_key
        fp_field = "%s_fp" % eval_key
        fn_field = "%s_fn" % eval_key

        dataset.add_sample_field(tp_field, fof.IntField)
        dataset.add_sample_field(fp_field, fof.IntField)
        dataset.add_sample_field(fn_field, fof.IntField)

        if processing_frames:
            dataset.add_frame_field(tp_field, fof.IntField)
            dataset.add_frame_field(fp_field, fof.IntField)
            dataset.add_frame_field(fn_field, fof.IntField)

        if not dynamic:
            return

        id_key = "%s_id" % eval_key
        iou_key = "%s_iou" % eval_key

        _, gt_prefix = samples._get_label_field_path(_gt_field)
        gt_prefix, _ = samples._handle_frame_field(gt_prefix)

        gt_eval = gt_prefix + "." + eval_key
        gt_id = gt_prefix + "." + id_key
        gt_iou = gt_prefix + "." + iou_key

        _, pred_prefix = samples._get_label_field_path(_pred_field)
        pred_prefix, _ = samples._handle_frame_field(pred_prefix)

        pred_eval = pred_prefix + "." + eval_key
        pred_id = pred_prefix + "." + id_key
        pred_iou = pred_prefix + "." + iou_key

        if processing_frames:
            dataset.add_frame_field(gt_eval, fof.StringField)
            dataset.add_frame_field(gt_id, fof.StringField)
            dataset.add_frame_field(gt_iou, fof.FloatField)
            dataset.add_frame_field(pred_eval, fof.StringField)
            dataset.add_frame_field(pred_id, fof.StringField)
            dataset.add_frame_field(pred_iou, fof.FloatField)
        else:
            dataset.add_sample_field(gt_eval, fof.StringField)
            dataset.add_sample_field(gt_id, fof.StringField)
            dataset.add_sample_field(gt_iou, fof.FloatField)
            dataset.add_sample_field(pred_eval, fof.StringField)
            dataset.add_sample_field(pred_id, fof.StringField)
            dataset.add_sample_field(pred_iou, fof.FloatField)

    def evaluate(self, doc, eval_key=None):
        """Evaluates the ground truth and predictions in the given document.

        Args:
            doc: a :class:`fiftyone.core.document.Document`
            eval_key (None): the evaluation key for this evaluation

        Returns:
            a list of matched ``(gt_label, pred_label, iou, pred_confidence)``
            tuples
        """
        raise NotImplementedError("subclass must implement evaluate()")

    def generate_results(
        self, samples, matches, eval_key=None, classes=None, missing=None
    ):
        """Generates aggregate evaluation results for the samples.

        Subclasses may perform additional computations here such as IoU sweeps
        in order to generate mAP, PR curves, etc.

        Args:
            samples: a :class:`fiftyone.core.collections.SampleCollection`
            matches: a list of
                ``(gt_label, pred_label, iou, pred_confidence, gt_id, pred_id)``
                matches. Either label can be ``None`` to indicate an unmatched
                object
            eval_key (None): the evaluation key for this evaluation
            classes (None): the list of possible classes. If not provided, the
                observed ground truth/predicted labels are used for results
                purposes
            missing (None): a missing label string. Any unmatched objects are
                given this label for results purposes

        Returns:
            a :class:`DetectionResults`
        """
        return DetectionResults(
            matches,
            eval_key=eval_key,
            gt_field=self.config.gt_field,
            pred_field=self.config.pred_field,
            classes=classes,
            missing=missing,
            samples=samples,
        )

    def get_fields(self, samples, eval_key):
        pred_field = self.config.pred_field
        pred_type = samples._get_label_field_type(pred_field)
        pred_key = "%s.%s.%s" % (
            pred_field,
            pred_type._LABEL_LIST_FIELD,
            eval_key,
        )

        gt_field = self.config.gt_field
        gt_type = samples._get_label_field_type(gt_field)
        gt_key = "%s.%s.%s" % (gt_field, gt_type._LABEL_LIST_FIELD, eval_key)

        fields = [
            "%s_tp" % eval_key,
            "%s_fp" % eval_key,
            "%s_fn" % eval_key,
            pred_key,
            "%s_id" % pred_key,
            "%s_iou" % pred_key,
            gt_key,
            "%s_id" % gt_key,
            "%s_iou" % gt_key,
        ]

        if samples._is_frame_field(gt_field):
            prefix = samples._FRAMES_PREFIX + eval_key
            fields.extend(
                ["%s_tp" % prefix, "%s_fp" % prefix, "%s_fn" % prefix]
            )

        return fields

    def cleanup(self, samples, eval_key):
        fields = [
            "%s_tp" % eval_key,
            "%s_fp" % eval_key,
            "%s_fn" % eval_key,
        ]

        try:
            pred_field, _ = samples._handle_frame_field(self.config.pred_field)
            pred_type = samples._get_label_field_type(self.config.pred_field)
            pred_key = "%s.%s.%s" % (
                pred_field,
                pred_type._LABEL_LIST_FIELD,
                eval_key,
            )
            fields.extend([pred_key, "%s_id" % pred_key, "%s_iou" % pred_key])
        except ValueError:
            # Field no longer exists, nothing to cleanup
            pass

        try:
            gt_field, _ = samples._handle_frame_field(self.config.gt_field)
            gt_type = samples._get_label_field_type(self.config.gt_field)
            gt_key = "%s.%s.%s" % (
                gt_field,
                gt_type._LABEL_LIST_FIELD,
                eval_key,
            )
            fields.extend([gt_key, "%s_id" % gt_key, "%s_iou" % gt_key])
        except ValueError:
            # Field no longer exists, nothing to cleanup
            pass

        if samples._is_frame_field(self.config.pred_field):
            samples._dataset.delete_sample_fields(
                ["%s_tp" % eval_key, "%s_fp" % eval_key, "%s_fn" % eval_key],
                error_level=1,
            )
            samples._dataset.delete_frame_fields(fields, error_level=1)
        else:
            samples._dataset.delete_sample_fields(fields, error_level=1)

    def _validate_run(self, samples, eval_key, existing_info):
        self._validate_fields_match(eval_key, "pred_field", existing_info)
        self._validate_fields_match(eval_key, "gt_field", existing_info)


class DetectionResults(BaseEvaluationResults):
    """Class that stores the results of a detection evaluation.

    Args:
        matches: a list of
            ``(gt_label, pred_label, iou, pred_confidence, gt_id, pred_id)``
            matches. Either label can be ``None`` to indicate an unmatched
            object
        eval_key (None): the evaluation key for this evaluation
        gt_field (None): the name of the ground truth field
        pred_field (None): the name of the predictions field
        classes (None): the list of possible classes. If not provided, the
            observed ground truth/predicted labels are used
        missing (None): a missing label string. Any unmatched objects are given
            this label for evaluation purposes
        samples (None): the :class:`fiftyone.core.collections.SampleCollection`
            for which the results were computed
    """

    def __init__(
        self,
        matches,
        eval_key=None,
        gt_field=None,
        pred_field=None,
        classes=None,
        missing=None,
        samples=None,
    ):
        if matches:
            ytrue, ypred, ious, confs, ytrue_ids, ypred_ids = zip(*matches)
        else:
            ytrue, ypred, ious, confs, ytrue_ids, ypred_ids = (
                [],
                [],
                [],
                [],
                [],
                [],
            )

        super().__init__(
            ytrue,
            ypred,
            confs=confs,
            eval_key=eval_key,
            gt_field=gt_field,
            pred_field=pred_field,
            ytrue_ids=ytrue_ids,
            ypred_ids=ypred_ids,
            classes=classes,
            missing=missing,
            samples=samples,
        )
        self.ious = np.array(ious)

    @classmethod
    def _from_dict(cls, d, samples, config, **kwargs):
        ytrue = d["ytrue"]
        ypred = d["ypred"]
        ious = d["ious"]

        confs = d.get("confs", None)
        if confs is None:
            confs = itertools.repeat(None)

        ytrue_ids = d.get("ytrue_ids", None)
        if ytrue_ids is None:
            ytrue_ids = itertools.repeat(None)

        ypred_ids = d.get("ypred_ids", None)
        if ypred_ids is None:
            ypred_ids = itertools.repeat(None)

        eval_key = d.get("eval_key", None)
        gt_field = d.get("gt_field", None)
        pred_field = d.get("pred_field", None)
        classes = d.get("classes", None)
        missing = d.get("missing", None)

        matches = list(zip(ytrue, ypred, ious, confs, ytrue_ids, ypred_ids))

        return cls(
            matches,
            eval_key=eval_key,
            gt_field=gt_field,
            pred_field=pred_field,
            classes=classes,
            missing=missing,
            samples=samples,
            **kwargs,
        )


def _parse_config(pred_field, gt_field, method, is_temporal, **kwargs):
    if method is None:
        if is_temporal:
            method = "activitynet"
        else:
            method = "coco"

    if method == "activitynet":
        from .activitynet import ActivityNetEvaluationConfig

        return ActivityNetEvaluationConfig(pred_field, gt_field, **kwargs)

    if method == "coco":
        from .coco import COCOEvaluationConfig

        return COCOEvaluationConfig(pred_field, gt_field, **kwargs)

    if method == "open-images":
        from .openimages import OpenImagesEvaluationConfig

        return OpenImagesEvaluationConfig(pred_field, gt_field, **kwargs)

    raise ValueError("Unsupported evaluation method '%s'" % method)


def _tally_matches(matches):
    tp = 0
    fp = 0
    fn = 0
    for match in matches:
        gt_label = match[0]
        pred_label = match[1]
        if gt_label is None:
            fp += 1
        elif pred_label is None:
            fn += 1
        elif gt_label != pred_label:
            fp += 1
            fn += 1
        else:
            tp += 1

    return tp, fp, fn



"""
FiftyOne Server /event route

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import typing as t

from starlette.endpoints import HTTPEndpoint
from starlette.requests import Request

import fiftyone.core.session.events as fose

from fiftyone.server.decorators import route
from fiftyone.server.events import dispatch_event


class Event(HTTPEndpoint):
    @route
    async def post(self, request: Request, data: t.Dict) -> t.Dict:
        await dispatch_event(
            data["subscription"],
            fose.Event.from_data(data["event"], data["data"]),
        )

        return {}



"""
Utilities for working with annotations in
`Scale AI format <https://docs.scale.com/reference#annotation>`_.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from collections import defaultdict
from copy import deepcopy
import logging
import os
from uuid import uuid4
import warnings

import numpy as np

import eta.core.image as etai
import eta.core.serial as etas
import eta.core.utils as etau
import eta.core.web as etaw

import fiftyone.core.labels as fol
import fiftyone.core.media as fomm
import fiftyone.core.metadata as fom
import fiftyone.core.utils as fou
import fiftyone.core.validation as fov


logger = logging.getLogger(__name__)


def import_from_scale(
    dataset,
    labels_dir_or_json,
    label_prefix=None,
    scale_id_field="scale_id",
):
    """Imports the Scale AI labels into the FiftyOne dataset.

    This method supports importing annotations from the following Scale API
    endpoints:

    -   `General Image Annotation <https://docs.scale.com/reference#general-image-annotation>`_
    -   `Semantic Segmentation Annotation <https://docs.scale.com/reference#semantic-segmentation-annotation>`_
    -   `General Video Annotation <https://docs.scale.com/reference#general-video-annotation>`_
    -   `Video Playback <https://docs.scale.com/reference#video-playback>`_

    The ``scale_id_field`` of the FiftyOne samples are used to associate
    samples with their corresponding Scale task IDs.

    The provided ``labels_dir_or_json`` can either be the path to a JSON
    export in the following format::

        [
            {
                "task_id": <scale-task-id1>,
                "response": {...},
                ...
            },
            {
                "task_id": <scale-task-id2>,
                "response": {...},
                ...
            },
            ...
        ]

    or a directory of per-task JSON files, which can either (a) directly
    contain the elements of the list above, or (b) contain task labels
    organized in the following format::

        labels_dir/
            <scale-task-id1>.json
            <scale-task-id2>.json
            ...

    where each JSON file contains only the contents of the ``response`` field
    for the task.

    The contents of the ``response`` field should be as follows:

    -   `General Image Annotation <https://docs.scale.com/reference#general-image-annotation>`_::

            {
                "annotations": [...]
                "global_attributes": {...}
            }

    -   `Semantic Segmentation Annotation <https://docs.scale.com/reference#semantic-segmentation-annotation>`_::

            {
                "annotations": {
                    ...
                    "combined": {
                        ...
                        "indexedImage": <url-or-filepath>
                    }
                },
                "labelMapping": {...}
            }

        where the ``indexedImage`` field (which is the only version of the
        segmentation used by this method) can contain either a URL, in which
        case the mask is downloaded from the web, or the path to the mask on
        disk.

    -   `General Video Annotation <https://docs.scale.com/reference#general-video-annotation>`_::

            {
                "annotations": {
                    "url": <url-or-filepath>
                },
                "events": {
                    "url": <url-or-filepath>
                }
            }

        where the ``url`` fields can contain either a URL, in which case the
        file is downloaded from the web, or the path to JSON file on disk.

        The annotations file should contain per-frame annotations in the
        following format::

            [
                {
                    "annotations": [...],
                    "global_attributes": {...}
                },
                {
                    "annotations": [...],
                    "global_attributes": {...}
                },
                ...
            ]

        where the n-th element in the list contains the labels for the n-th
        frame that was labeled.

        Note that, if parameters such as ``duration_time``, ``frame_rate``, and
        ``start_time`` were used to specify which frames of the video to
        annotate, then you must ensure that the ``labels_dir_or_json`` JSON
        that you provide to this method contains the ``task`` fields for each
        Scale task so that the correct frame numbers can be determined from
        these values.

        The optional events file should contain a list of events in the video::

            {
                "events": [...]
            }

    -   `Video Playback <https://docs.scale.com/reference#video-playback>`_::

            {
                "annotations": {
                    "url": <url-or-filepath>
                },
                "events": {
                    "url": <url-or-filepath>
                }
            }

        where the ``url`` fields can contain either a URL, in which case the
        file is downloaded from the web, or the path to JSON files on disk.

        The annotations file should contain a dictionary of object
        trajectories::

            {
                "annotations": {...}
            }

        The optional events file should contain a list of events in the video::

            {
                "events": [...]
            }

    Args:
        dataset: a :class:`fiftyone.core.dataset.Dataset`
        labels_dir_or_json: the path to a Scale AI JSON export or a directory
            of JSON exports as per the formats described above
        label_prefix (None): a prefix to prepend to the sample label field(s)
            that are created, separated by an underscore
        scale_id_field ("scale_id"): the sample field to use to associate Scale
            task IDs with FiftyOne samples
    """
    fov.validate_collection(dataset, media_type=(fomm.IMAGE, fomm.VIDEO))
    is_video = dataset.media_type == fomm.VIDEO

    # Load labels
    if labels_dir_or_json.endswith(".json"):
        labels = _load_labels(labels_dir_or_json)
    else:
        labels = _load_labels_dir(labels_dir_or_json)

    id_map = {k: v for k, v in zip(*dataset.values([scale_id_field, "id"]))}

    if label_prefix:
        label_key = lambda k: label_prefix + "_" + k
    else:
        label_key = lambda k: k

    with fou.ProgressBar(total=len(labels)) as pb:
        for task_id, task_labels in pb(labels.items()):
            if task_id not in id_map:
                logger.info(
                    "Skipping labels for unknown Scale ID '%s'", task_id
                )
                continue

            sample = dataset[id_map[task_id]]

            if sample.metadata is None:
                if is_video:
                    sample.metadata = fom.VideoMetadata.build_for(
                        sample.filepath
                    )
                else:
                    sample.metadata = fom.ImageMetadata.build_for(
                        sample.filepath
                    )

            if is_video:
                frames = _parse_video_labels(task_labels, sample.metadata)
                sample.frames.merge(
                    {
                        frame_number: {
                            label_key(fname): flabel
                            for fname, flabel in frame_dict.items()
                        }
                        for frame_number, frame_dict in frames.items()
                    }
                )
            else:
                frame_size = (sample.metadata.width, sample.metadata.height)
                anno_dict = task_labels["response"]
                labels_dict = _parse_image_labels(anno_dict, frame_size)
                sample.update_fields(
                    {label_key(k): v for k, v in labels_dict.items()}
                )

            sample.save()


# @todo add support for `duration_time`, `frame_rate`, and `start_time`
# parameters when exporting general video annotation labels?
def export_to_scale(
    sample_collection,
    json_path,
    video_labels_dir=None,
    video_events_dir=None,
    video_playback=False,
    label_field=None,
    frame_labels_field=None,
):
    """Exports labels from the FiftyOne samples to Scale AI format.

    This function is useful for generating pre-annotations that can be provided
    to Scale AI via the ``hypothesis`` parameter of the following endpoints:

    -   `General Image Annotation <https://docs.scale.com/reference#general-image-annotation>`_
    -   `General Video Annotation <https://docs.scale.com/reference#general-video-annotation>`_
    -   `Video Playback <https://docs.scale.com/reference#video-playback>`_

    The output ``json_path`` will be a JSON file in the following format::

        {
            <sample-id1>: {
                "filepath": <filepath1>,
                "hypothesis": {...}
            },
            <sample-id2>: {
                "filepath": <filepath2>,
                "hypothesis": {...}
            },
            ...
        }

    The format of the ``hypothesis`` field depends on the label type:

    -   Sample-level classifications, detections, polylines, polygons, and
        keypoints::

            {
                "annotations": [...],
                "global_attributes": {...}
            }

    -   Video samples::

            {
                "annotations": {
                    "url": <filepath>
                }
            }

    When exporting labels for video datasets, the ``url`` field will contain
    the paths on disk to per-sample JSON files that are written to
    ``video_labels_dir`` as follows::

        video_labels_dir/
            <sample-id1>.json
            <sample-id2>.json
            ...

    When ``video_playback == False``, the per-sample JSON files are written in
    `General Video Annotation format <https://docs.scale.com/reference#general-video-annotation>`_::

        [
            {
                "annotations": [...],
                "global_attributes": {...}
            },
            {
                "annotations": [...],
                "global_attributes": {...}
            },
            ...
        ]

    where the n-th element in the list contains the labels for the n-th frame
    of the video.

    When ``video_playback == True``, the per-sample JSON files are written in
    `Video Playback format <https://docs.scale.com/reference#video-playback>`_::

        {
            "annotations": {...}
        }

    When exporting labels for videos and the ``video_events_dir`` parameter is
    provided, the ``hypothesis`` fields of the JSON written to ``json_path``
    will include an ``events`` field::

        {
            "annotations": {
                "url": <filepath>
            },
            "events": {
                "url": <filepath>
            }
        }

    whose ``url`` field will contain the paths on disk to per-sample JSON files
    that are written to ``video_events_dir`` as follows::

        video_events_dir/
            <sample-id1>.json
            <sample-id2>.json
            ...

    where each per-sample JSON file contains the
    `events in the video <https://docs.scale.com/reference#events>`_::

        {
            "events": [...]
        }

    Args:
        sample_collection: a
            :class:`fiftyone.core.collections.SampleCollection`
        json_path: the path to write the JSON export
        video_labels_dir (None): a directory to write the per-sample video
            labels. Only applicable for video datasets
        video_events_dir (None): a directory to write the per-sample video
            events. Only applicable for video datasets
        video_playback (False): whether to export video labels in a suitable
            format for use with the
            `Video Playback <https://docs.scale.com/reference#video-playback>`_ task.
            By default, video labels are exported for in a suitable format for
            the `General Video Annotation <https://docs.scale.com/reference#general-video-annotation>`_
            task. Only applicable for video datasets
        label_field (None): optional label field(s) to export. Can be any of
            the following:

            -   the name of a label field to export
            -   a glob pattern of label field(s) to export
            -   a list or tuple of label field(s) to export
            -   a dictionary mapping label field names to keys to use when
                constructing the exported labels

            By default, no labels are exported
        frame_labels_field (None): optional frame label field(s) to export.
            Only applicable to video datasets. Can be any of the following:

            -   the name of a frame label field to export
            -   a glob pattern of frame label field(s) to export
            -   a list or tuple of frame label field(s) to export
            -   a dictionary mapping frame label field names to keys to use
                when constructing the exported frame labels

            By default, no frame labels are exported
    """
    fov.validate_collection(
        sample_collection, media_type=(fomm.IMAGE, fomm.VIDEO)
    )
    is_video = sample_collection.media_type == fomm.VIDEO

    # Get label fields to export
    label_fields = sample_collection._parse_label_field(
        label_field,
        allow_coercion=False,
        force_dict=True,
        required=False,
    )

    # Get frame label fields to export
    if is_video:
        frame_label_fields = sample_collection._parse_frame_labels_field(
            frame_labels_field,
            allow_coercion=False,
            required=False,
            force_dict=True,
        )

        if frame_label_fields and (
            video_labels_dir is None and video_events_dir is None
        ):
            raise ValueError(
                "Must provide `video_labels_dir` and/or `video_events_dir` "
                "when exporting labels for video datasets"
            )

    # Compute metadata if necessary
    sample_collection.compute_metadata()

    # Export the labels
    labels = {}
    for sample in sample_collection.iter_samples(progress=True):
        metadata = sample.metadata

        # Get frame size
        if is_video:
            frame_size = (metadata.frame_width, metadata.frame_height)
        else:
            frame_size = (metadata.width, metadata.height)

        # Export sample-level labels
        if label_fields:
            labels_dict = _get_labels(sample, label_fields)
            anno_dict = _to_scale_image_labels(labels_dict, frame_size)

        # Export frame-level labels
        if is_video and frame_label_fields:
            frames = _get_frame_labels(sample, frame_label_fields)
            make_events = video_events_dir is not None
            if video_playback:
                annotations, events = _to_scale_video_playback_labels(
                    frames, frame_size, make_events=make_events
                )
            else:
                annotations, events = _to_scale_video_annotation_labels(
                    frames, frame_size, make_events=make_events
                )

            anno_dict = {}

            # Write annotations
            if video_labels_dir:
                anno_path = os.path.join(video_labels_dir, sample.id + ".json")
                etas.write_json(annotations, anno_path)
                anno_dict["annotations"] = {"url": anno_path}

            # Write events
            if video_events_dir:
                events_path = os.path.join(
                    video_events_dir, sample.id + ".json"
                )
                etas.write_json(events, events_path)
                anno_dict["events"] = {"url": events_path}

        labels[sample.id] = {
            "filepath": sample.filepath,
            "hypothesis": anno_dict,
        }

    etas.write_json(labels, json_path)


def convert_scale_export_to_import(inpath, outpath):
    """Converts a Scale AI JSON export generated by :meth:`export_to_scale`
    into the format expected by :meth:`import_from_scale`.

    The output JSON file will have the same format that is generated when
    performing a bulk export of a Scale AI project's labels.

    Args:
        inpath: the path to an JSON file generated (for example) by
            :meth:`export_to_scale`
        outpath: the path to write a JSON file containing the converted labels

    Returns:
        a dictionary mapping sample IDs to the task IDs that were generated for
        each sample
    """
    labels = etas.read_json(inpath)

    id_map = {}
    annos = []
    for sample_id, task_dict in labels.items():
        task_id = str(uuid4())
        id_map[sample_id] = task_id
        annos.append({"task_id": task_id, "response": task_dict["hypothesis"]})

    etas.write_json(annos, outpath)

    return id_map


def _load_labels(json_path):
    d_list = etas.read_json(json_path)
    return {d["task_id"]: d for d in d_list}


def _load_labels_dir(labels_dir):
    labels = {}
    json_patt = os.path.join(labels_dir, "*.json")
    for json_path in etau.get_glob_matches(json_patt):
        task_id, task_labels = _load_task_labels(json_path)
        labels[task_id] = task_labels

    return labels


def _load_task_labels(json_path):
    d = etas.read_json(json_path)

    if "task_id" in d:
        task_id = d["task_id"]
        labels = d
    else:
        task_id = os.path.splitext(os.path.basename(json_path))[0]
        labels = {
            "task_id": task_id,
            "response": d,
            "task": None,
        }

    return task_id, labels


def _get_labels(sample_or_frame, label_fields):
    labels_dict = {}
    for field, key in label_fields.items():
        value = sample_or_frame[field]
        if value is not None:
            labels_dict[key] = value

    return labels_dict


def _get_frame_labels(sample, frame_label_fields):
    frames = {}
    for frame_number, frame in sample.frames.items():
        frames[frame_number] = _get_labels(frame, frame_label_fields)

    return frames


# https://docs.scale.com/reference#general-image-annotation
def _to_scale_image_labels(labels_dict, frame_size):
    annotations = []
    global_attributes = {}
    for name, label in labels_dict.items():
        if isinstance(label, fol.Classification):
            attrs_dict = _to_classification(name, label)
            global_attributes.update(attrs_dict)
        elif isinstance(label, (fol.Detection, fol.Detections)):
            annos = _to_detections(label, frame_size)
            annotations.extend(annos)
        elif isinstance(label, (fol.Polyline, fol.Polylines)):
            annos = _to_polylines(label, frame_size)
            annotations.extend(annos)
        elif isinstance(label, (fol.Keypoint, fol.Keypoints)):
            annos = _to_points(label, frame_size)
            annotations.extend(annos)
        elif label is not None:
            msg = "Ignoring unsupported label type '%s'" % label.__class__
            warnings.warn(msg)

    anno_dict = {"annotations": annotations}

    if global_attributes:
        anno_dict["global_attributes"] = global_attributes

    return anno_dict


# https://docs.scale.com/reference#general-video-annotation
def _to_scale_video_annotation_labels(frames, frame_size, make_events=False):
    in_progress_events = {}
    events = []

    annotations = []
    for frame_number, labels_dict in frames.items():
        # Generate events, if requested
        if make_events:
            _parse_frame_events(
                labels_dict, frame_number, events, in_progress_events
            )

        # Generate frame labels
        anno_dict = _to_scale_image_labels(labels_dict, frame_size)
        annotations.append(anno_dict)

    # Finalize all remaining events
    _finalize_events(events, None, in_progress_events)

    return annotations, {"events": events}


# https://docs.scale.com/reference#video-playback
def _to_scale_video_playback_labels(frames, frame_size, make_events=False):
    in_progress_events = {}
    events = []

    traj_uuids = {}
    annotations = {}
    for frame_number, frame in frames.items():
        # Ingst new labels
        for label in frame.values():
            if make_events and isinstance(
                label, (fol.Classification, fol.Classifications)
            ):
                _ingest_event_label(label, frame_number, in_progress_events)
            elif isinstance(label, (fol.Detection, fol.Detections)):
                if isinstance(label, fol.Detection):
                    detections = [label]
                else:
                    detections = label.detections

                for detection in detections:
                    if detection.index is None:
                        key = None
                    else:
                        key = (detection.label, detection.index)

                    uuid = traj_uuids.get(key, None)
                    if uuid is None:
                        uuid = str(uuid4())
                        annotations[uuid] = _init_video_box(detection.label)

                        if key is not None:
                            traj_uuids[key] = uuid

                    annotations[uuid]["frames"].append(
                        _make_video_box_frame(
                            detection, frame_number, frame_size
                        )
                    )

            elif label is not None:
                msg = "Ignoring unsupported label type '%s'" % label.__class__
                warnings.warn(msg)

        # Finalize in-progress events
        _finalize_events(events, frame_number, in_progress_events)

    # Finalize all remaining events
    _finalize_events(events, None, in_progress_events)

    return {"annotations": annotations}, {"events": events}


def _parse_frame_events(labels_dict, frame_number, events, in_progress_events):
    for name, label in labels_dict.items():
        if isinstance(label, (fol.Classification, fol.Classifications)):
            _ingest_event_label(label, frame_number, in_progress_events)
            labels_dict[name] = None

    # Finalize in-progress events
    _finalize_events(events, frame_number, in_progress_events)


def _ingest_event_label(label, frame_number, in_progress_events):
    if isinstance(label, fol.Classification):
        classifications = [label]
    else:
        classifications = label.classifications

    for classification in classifications:
        event_label = classification.label
        if event_label in in_progress_events:
            start = in_progress_events[event_label][0]
        else:
            start = frame_number

        in_progress_events[event_label] = (start, frame_number)


def _finalize_events(events, frame_number, in_progress_events):
    for event_label in list(in_progress_events.keys()):
        start, last = in_progress_events[event_label]
        if last != frame_number:
            events.append(_make_event(event_label, start, last))
            del in_progress_events[event_label]


# https://docs.scale.com/reference#events
def _make_event(label, start, end):
    if end == start:
        return {
            "label": label,
            "type": "point",
            "start": start,
        }

    return {
        "label": label,
        "type": "range",
        "start": start,
        "end": end,
    }


# https://docs.scale.com/reference#video-playback
def _init_video_box(label):
    return {
        "label": label,
        "geometry": "box",
        "frames": [],
    }


# https://docs.scale.com/reference#video-playback
def _make_video_box_frame(detection, frame_number, frame_size):
    x, y, w, h = detection.bounding_box
    width, height = frame_size
    frame_dict = {
        "key": frame_number,
        "x": round(x * width, 1),
        "y": round(y * height, 1),
        "width": round(w * width, 1),
        "height": round(h * height, 1),
    }

    attributes = _get_attributes(detection)
    if attributes:
        frame_dict["attributes"] = attributes

    return frame_dict


# https://docs.scale.com/reference#global-attributes
def _to_classification(name, label):
    return {name: label.label}


# https://docs.scale.com/reference#attributes-overview
def _get_attributes(label):
    attrs = {}
    for name, value in label.iter_attributes():
        if etau.is_str(value) or etau.is_numeric(value):
            attrs[name] = value
        else:
            msg = "Ignoring unsupported attribute type '%s'" % type(value)
            warnings.warn(msg)

    return attrs


# https://docs.scale.com/reference#boxes
def _to_detections(label, frame_size):
    if isinstance(label, fol.Detections):
        detections = label.detections
    else:
        detections = [label]

    annos = []
    for detection in detections:
        anno = _make_base_anno("box", detection.label)
        anno.update(_make_bbox(detection.bounding_box, frame_size))

        attributes = _get_attributes(detection)
        if attributes:
            anno["attributes"] = attributes

        annos.append(anno)

    return annos


# https://docs.scale.com/reference#polygons
# https://docs.scale.com/reference#line
def _to_polylines(label, frame_size):
    if isinstance(label, fol.Polylines):
        polylines = label.polylines
    else:
        polylines = [label]

    annos = []
    for polyline in polylines:
        type_ = "polygon" if polyline.filled else "line"
        attributes = _get_attributes(polyline)
        for points in polyline.points:
            anno = _make_base_anno(type_, polyline.label)
            anno["vertices"] = [_make_point(p, frame_size) for p in points]
            if attributes:
                anno["attributes"] = attributes

            annos.append(anno)

    return annos


# https://docs.scale.com/reference#points
def _to_points(label, frame_size):
    if isinstance(label, fol.Keypoints):
        keypoints = label.keypoints
    else:
        keypoints = [keypoints]

    annos = []
    for keypoint in keypoints:
        attributes = _get_attributes(keypoint)
        for point in keypoint.points:
            anno = _make_base_anno("point", keypoint.label)
            anno.update(_make_point(point, frame_size))
            if attributes:
                anno["attributes"] = attributes

            annos.append(anno)

    return annos


def _make_base_anno(type_, label):
    return {
        "type": type_,
        "label": label,
        "uuid": str(uuid4()),
    }


def _make_bbox(bounding_box, frame_size):
    x, y, w, h = bounding_box
    width, height = frame_size
    return {
        "left": round(x * width, 1),
        "top": round(y * height, 1),
        "width": round(w * width, 1),
        "height": round(h * height, 1),
    }


def _make_point(point, frame_size):
    x, y = point
    width, height = frame_size
    return {"x": round(x * width, 1), "y": round(y * height, 1)}


def _parse_video_labels(task_labels, metadata):
    anno_dict = task_labels["response"]

    if "annotations" in anno_dict:
        annos = _download_or_load_json(anno_dict["annotations"]["url"])
    else:
        annos = None

    if "events" in anno_dict:
        events = _download_or_load_json(anno_dict["events"]["url"])
    else:
        events = None

    if isinstance(annos, dict):
        return _parse_video_playback_labels(annos, events, metadata)

    task = task_labels.get("task", None)
    return _parse_video_annotation_labels(annos, events, metadata, task=task)


# https://docs.scale.com/reference#general-video-annotation
def _parse_video_annotation_labels(annos, events, metadata, task=None):
    if annos is not None:
        annos = annos.get("annotations", [])
    else:
        annos = []

    frame_numbers = None
    if task is not None:
        try:
            frame_numbers = _get_frame_numbers(task, metadata)
        except:
            pass

    if frame_numbers is None:
        frame_numbers = range(1, len(annos) + 1)

    if len(frame_numbers) != len(annos):
        logger.warning(
            "Unable to determine which frame numbers have labels. You may "
            "need to inspect the labels that are loaded..."
        )

    frame_size = (metadata.frame_width, metadata.frame_height)

    frames = defaultdict(dict)

    # Parse frame labels
    for frame_number, anno_dict in zip(frame_numbers, annos):
        frames[frame_number] = _parse_image_labels(anno_dict, frame_size)

    # Parse events
    if events is not None:
        events = events.get("events", [])
        _parse_video_playback_events(frames, events)

    return frames


def _get_frame_numbers(task, metadata):
    params = task["params"]

    if params["attachment_type"] != "video":
        return None

    frame_rate = metadata.frame_rate

    start_time = params.get("start_time", None)
    if start_time is not None:
        first_frame = 1 + int(round(frame_rate * start_time))
    else:
        first_frame = 1

    duration = params.get("duration_time", None)
    if duration is not None:
        last_frame = 1 + int(round(frame_rate * duration))
    else:
        last_frame = metadata.total_frame_count

    sampling_rate = params.get("frame_rate", frame_rate)
    accel = frame_rate / sampling_rate

    return sorted(
        set(int(round(f)) for f in np.arange(first_frame, last_frame, accel))
    )


# https://docs.scale.com/reference#video-playback
def _parse_video_playback_labels(annos, events, metadata):
    frames = defaultdict(dict)

    # Parse events
    if events is not None:
        events = events.get("events", [])
        _parse_video_playback_events(frames, events)

    # Parse video objects
    if annos is not None:
        annotations = annos.get("annotations", {})
        _parse_video_playback_objects(frames, annotations, metadata)

    return frames


def _parse_video_playback_events(frames, events):
    events_map = defaultdict(list)
    for event in events:
        type_ = event["type"]
        label = event["label"]
        start = event["start"]
        if type_ == "range":
            end = event["end"]
        elif type_ == "point":
            end = start
        else:
            msg = "Ignoring unsupported event type '%s'" % type_
            warnings.warn(msg)
            continue

        attributes = _parse_classifications(event.get("attributes", {}))

        for frame_number in range(start, end + 1):
            events_map[frame_number].append(fol.Classification(label=label))
            if attributes:
                frames[frame_number].update(deepcopy(attributes))

    for frame_number, classifications in events_map.items():
        frames[frame_number]["events"] = fol.Classifications(
            classifications=classifications
        )


def _parse_video_playback_objects(frames, annotations, metadata):
    frame_size = (metadata.frame_width, metadata.frame_height)

    index = 0
    detections_map = defaultdict(list)
    for _, video_obj in annotations.items():
        index += 1
        label = video_obj["label"]
        type_ = video_obj["geometry"]
        if type_ != "box":
            msg = "Ignoring unsupported video annotation geometry '%s'" % type_
            warnings.warn(msg)
            continue

        for frame_dict in video_obj["frames"]:
            frame_number = frame_dict["key"]
            attributes = frame_dict.get("attributes", {})
            bounding_box = _parse_video_bbox(frame_dict, frame_size)
            detection = fol.Detection(
                label=label,
                index=index,
                bounding_box=bounding_box,
                **attributes,
            )
            detections_map[frame_number].append(detection)

    for frame_number, detections in detections_map.items():
        frames[frame_number]["detections"] = fol.Detections(
            detections=detections
        )


# https://docs.scale.com/reference#general-image-annotation
def _parse_image_labels(anno_dict, frame_size):
    labels = {}

    # Parse classifications
    attrs = anno_dict.get("global_attributes", {})
    classifications = _parse_classifications(attrs)
    labels.update(classifications)

    # Parse objects
    objects = _parse_objects(anno_dict, frame_size)
    labels.update(objects)

    return labels


def _parse_classifications(attrs):
    return {k: fol.Classification(label=v) for k, v in attrs.items()}


def _parse_objects(anno_dict, frame_size):
    annos = anno_dict.get("annotations", [])

    if isinstance(annos, dict):
        return {"segmentation": _parse_mask(anno_dict)}

    detections = []
    polylines = []
    keypoints = []
    for anno in annos:
        type_ = anno["type"]
        label = anno.get("label", None)  # Scale supports unlabeled tasks

        attributes = anno.get("attributes", {})

        if type_ == "box":
            # Detection
            bounding_box = _parse_bbox(anno, frame_size)
            detections.append(
                fol.Detection(
                    label=label, bounding_box=bounding_box, **attributes
                )
            )
        elif type_ == "polygon":
            # Polyline
            points = _parse_points(anno["vertices"], frame_size)
            polylines.append(
                fol.Polyline(
                    label=label,
                    points=[points],
                    closed=True,
                    filled=True,
                    **attributes,
                )
            )
        elif type_ == "line":
            # Polyline
            points = _parse_points(anno["vertices"], frame_size)
            polylines.append(
                fol.Polyline(
                    label=label,
                    points=[points],
                    closed=True,
                    filled=False,
                    **attributes,
                )
            )
        elif type_ == "point":
            # Polyline
            point = _parse_point(anno, frame_size)
            keypoints.append(
                fol.Keypoint(label=label, points=[point], **attributes)
            )
        else:
            msg = "Ignoring unsupported label type '%s'" % type_
            warnings.warn(msg)

    labels = {}

    if detections:
        labels["detections"] = fol.Detections(detections=detections)

    if polylines:
        labels["polylines"] = fol.Polylines(polylines=polylines)

    if keypoints:
        labels["keypoints"] = fol.Keypoints(keypoints=keypoints)

    return labels


def _parse_bbox(anno, frame_size):
    width, height = frame_size
    x = anno["left"] / width
    y = anno["top"] / height
    w = anno["width"] / width
    h = anno["height"] / height
    return [x, y, w, h]


def _parse_video_bbox(anno, frame_size):
    width, height = frame_size
    x = anno["x"] / width
    y = anno["y"] / height
    w = anno["width"] / width
    h = anno["height"] / height
    return [x, y, w, h]


def _parse_points(vertices, frame_size):
    return [_parse_point(vertex, frame_size) for vertex in vertices]


def _parse_point(vertex, frame_size):
    width, height = frame_size
    return (vertex["x"] / width, vertex["y"] / height)


# https://docs.scale.com/reference#segmentannotation-callback-format
def _parse_mask(anno_dict):
    indexed_image_uri = anno_dict["annotations"]["combined"]["indexedImage"]
    mask = _download_or_load_image(indexed_image_uri)

    segmentation = fol.Segmentation(mask=mask)

    label_mapping = anno_dict.get("labelMapping", None)
    if label_mapping is not None:
        label_map = {}
        for label, value in label_mapping.items():
            if isinstance(value, dict):
                value = [value]

            for instance in value:
                label_map[instance["index"]] = label

        segmentation.label_map = label_map

    return segmentation


def _download_or_load_json(url_or_filepath):
    if url_or_filepath.startswith("http"):
        json_bytes = etaw.download_file(url_or_filepath, quiet=True)
        return etas.load_json(json_bytes)

    return etas.read_json(url_or_filepath)


def _download_or_load_image(url_or_filepath):
    if url_or_filepath.startswith("http"):
        img_bytes = etaw.download_file(url_or_filepath, quiet=True)
        return etai.decode(img_bytes)

    return etai.read(url_or_filepath)



"""
FiftyOne Server GraphQL scalars

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from datetime import date, datetime, timezone
from bson import json_util
import json
import strawberry as gql
import typing as t

from fiftyone.core.json import stringify
from fiftyone.core.utils import datetime_to_timestamp, timestamp_to_datetime


BSON = gql.scalar(
    t.NewType("BSON", object),
    serialize=lambda v: json.loads(json_util.dumps(v)),
    parse_value=lambda v: json_util.loads(json.dumps(v)),
)

BSONArray = gql.scalar(
    t.NewType("BSONArray", object),
    serialize=lambda v: json.loads(json_util.dumps(v)),
    parse_value=lambda v: json_util.loads(json.dumps(v)),
)

JSON = gql.scalar(
    t.NewType("JSON", object),
    serialize=lambda v: stringify(v),
    parse_value=lambda v: v,
)

JSONArray = gql.scalar(
    t.NewType("JSONArray", object),
    serialize=lambda v: json.loads(json_util.dumps(v)),
    parse_value=lambda v: json_util.loads(json.dumps(v)),
)

DateTime = gql.scalar(
    datetime,
    serialize=lambda v: datetime_to_timestamp(v),
    parse_value=lambda v: timestamp_to_datetime(v),
)


Date = gql.scalar(
    date,
    serialize=lambda v: datetime_to_timestamp(v),
    parse_value=lambda v: timestamp_to_datetime(v),
)



"""
Regression evaluation.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import logging
import itertools
import numbers

import numpy as np
import sklearn.metrics as skm
from tabulate import tabulate

import eta.core.utils as etau

import fiftyone.core.evaluation as foe
import fiftyone.core.fields as fof
import fiftyone.core.labels as fol
import fiftyone.core.plots as fop
import fiftyone.core.validation as fov


logger = logging.getLogger(__name__)


def evaluate_regressions(
    samples,
    pred_field,
    gt_field="ground_truth",
    eval_key=None,
    missing=None,
    method="simple",
    **kwargs,
):
    """Evaluates the regression predictions in the given collection with
    respect to the specified ground truth values.

    You can customize the evaluation method by passing additional
    parameters for the method's config class as ``kwargs``.

    The supported ``method`` values and their associated configs are:

    -   ``"simple"``: :class:`SimpleEvaluationConfig`

    If an ``eval_key`` is specified, then this method will record some
    statistics on each sample:

    -   When evaluating sample-level fields, an ``eval_key`` field will be
        populated on each sample recording the error of that sample's
        prediction.

    -   When evaluating frame-level fields, an ``eval_key`` field will be
        populated on each frame recording the error of that frame's
        prediction. In addition, an ``eval_key`` field will be populated on
        each sample that records the average error of the frame predictions
        of the sample.

    Args:
        samples: a :class:`fiftyone.core.collections.SampleCollection`
        pred_field: the name of the field containing the predicted
            :class:`fiftyone.core.labels.Regression` instances
        gt_field ("ground_truth"): the name of the field containing the
            ground truth :class:`fiftyone.core.labels.Regression` instances
        eval_key (None): a string key to use to refer to this evaluation
        missing (None): a missing value. Any None-valued regressions are
            given this value for results purposes
        method ("simple"): a string specifying the evaluation method to use.
            Supported values are ``("simple")``
        **kwargs: optional keyword arguments for the constructor of the
            :class:`RegressionEvaluationConfig` being used

    Returns:
        a :class:`RegressionResults`
    """
    fov.validate_collection_label_fields(
        samples, (pred_field, gt_field), fol.Regression, same_type=True
    )

    config = _parse_config(pred_field, gt_field, method, **kwargs)
    eval_method = config.build()
    eval_method.ensure_requirements()

    eval_method.register_run(samples, eval_key)
    eval_method.register_samples(samples, eval_key)

    results = eval_method.evaluate_samples(
        samples, eval_key=eval_key, missing=missing
    )
    eval_method.save_run_results(samples, eval_key, results)

    return results


class RegressionEvaluationConfig(foe.EvaluationMethodConfig):
    """Base class for configuring :class:`RegressionEvaluation` instances.

    Args:
        pred_field: the name of the field containing the predicted
            :class:`fiftyone.core.labels.Regression` instances
        gt_field ("ground_truth"): the name of the field containing the ground
            truth :class:`fiftyone.core.labels.Regression` instances
    """

    def __init__(self, pred_field, gt_field, **kwargs):
        super().__init__(**kwargs)
        self.pred_field = pred_field
        self.gt_field = gt_field


class RegressionEvaluation(foe.EvaluationMethod):
    """Base class for regression evaluation methods.

    Args:
        config: a :class:`RegressionEvaluationConfig`
    """

    def register_samples(self, samples, eval_key):
        """Registers the collection on which evaluation will be performed.

        This method will be called before calling :meth:`evaluate_samples`.
        Subclasses can extend this method to perform any setup required for an
        evaluation run.

        Args:
            samples: a :class:`fiftyone.core.collections.SampleCollection`
            eval_key: the evaluation key for this evaluation
        """
        if eval_key is None:
            return

        dataset = samples._dataset
        is_frame_field = samples._is_frame_field(self.config.gt_field)

        if is_frame_field:
            dataset.add_sample_field(eval_key, fof.FloatField)
            dataset.add_frame_field(eval_key, fof.FloatField)
        else:
            dataset.add_sample_field(eval_key, fof.FloatField)

    def evaluate_samples(self, samples, eval_key=None, missing=None):
        """Evaluates the regression predictions in the given samples with
        respect to the specified ground truth values.

        Args:
            samples: a :class:`fiftyone.core.collections.SampleCollection`
            eval_key (None): an evaluation key for this evaluation
            missing (None): a missing value. Any None-valued regressions are
                given this value for results purposes

        Returns:
            a :class:`RegressionResults` instance
        """
        pass

    def get_fields(self, samples, eval_key):
        fields = [eval_key]

        if samples._is_frame_field(self.config.gt_field):
            prefix = samples._FRAMES_PREFIX + eval_key
            fields.append(prefix)

        return fields

    def cleanup(self, samples, eval_key):
        fields = [eval_key]

        samples._dataset.delete_sample_fields(fields, error_level=1)
        if samples._is_frame_field(self.config.gt_field):
            samples._dataset.delete_frame_fields(fields, error_level=1)

    def _validate_run(self, samples, eval_key, existing_info):
        self._validate_fields_match(eval_key, "pred_field", existing_info)
        self._validate_fields_match(eval_key, "gt_field", existing_info)


class SimpleEvaluationConfig(RegressionEvaluationConfig):
    """Base class for configuring :class:`SimpleEvaluation` instances.

    Args:
        pred_field: the name of the field containing the predicted
            :class:`fiftyone.core.labels.Regression` instances
        gt_field: the name of the field containing the ground truth
            :class:`fiftyone.core.labels.Regression` instances
        metric ("squared_error"): the error metric to use to populate
            sample/frame-level error data. Supported values are
            ``("squared_error", "absolute_error")`` or any function that
            accepts two scalar arguments ``(ypred, ytrue)``
    """

    def __init__(self, pred_field, gt_field, metric="squared_error", **kwargs):
        super().__init__(pred_field, gt_field, **kwargs)
        self._metric = metric

    @property
    def method(self):
        return "simple"

    @property
    def metric(self):
        return self._metric if etau.is_str(self._metric) else "custom"

    def attributes(self):
        return super().attributes() + ["metric"]


class SimpleEvaluation(RegressionEvaluation):
    """Simple regression evaluation.

    Args:
        config: a :class:`SimpleEvaluationConfig`
    """

    def evaluate_samples(self, samples, eval_key=None, missing=None):
        metric = self.config._metric

        if metric == "squared_error":
            error_fcn = lambda yp, yt: (yp - yt) ** 2
        elif metric == "absolute_error":
            error_fcn = lambda yp, yt: abs(yp - yt)
        elif callable(metric):
            error_fcn = metric
        else:
            raise ValueError(
                "Unsupported metric '%s'. The supported values are %s or a "
                "function that accepts two scalar arguments `(ypred, ytrue)`"
                % (metric, ("squared_error", "absolute_error"))
            )

        pred_field = self.config.pred_field
        gt_field = self.config.gt_field
        is_frame_field = samples._is_frame_field(gt_field)

        gt = gt_field + ".value"
        pred = pred_field + ".value"
        pred_conf = pred_field + ".confidence"
        _id = "id" if not is_frame_field else "frames.id"

        ytrue, ypred, confs, ids = samples.values([gt, pred, pred_conf, _id])

        if is_frame_field:
            _ytrue = list(itertools.chain.from_iterable(ytrue))
            _ypred = list(itertools.chain.from_iterable(ypred))
            _confs = list(itertools.chain.from_iterable(confs))
            _ids = list(itertools.chain.from_iterable(ids))
        else:
            _ytrue = ytrue
            _ypred = ypred
            _confs = confs
            _ids = ids

        results = RegressionResults(
            _ytrue,
            _ypred,
            confs=_confs,
            eval_key=eval_key,
            gt_field=gt_field,
            pred_field=pred_field,
            ids=_ids,
            missing=missing,
            samples=samples,
        )

        if eval_key is None:
            return results

        def compute_error(yp, yt):
            if missing is not None:
                if yp is None:
                    yp = missing

                if yt is None:
                    yt = missing

            try:
                return error_fcn(yp, yt)
            except:
                return None

        if is_frame_field:
            frame_errors = [
                list(map(compute_error, yp, yt))
                for yp, yt in zip(ypred, ytrue)
            ]
            sample_errors = [_safe_mean(e) for e in frame_errors]

            eval_frame = samples._FRAMES_PREFIX + eval_key

            # Sample-level errors
            samples.set_values(eval_key, sample_errors)

            # Per-frame errors
            samples.set_values(eval_frame, frame_errors)
        else:
            errors = list(map(compute_error, ypred, ytrue))

            # Per-sample errors
            samples.set_values(eval_key, errors)

        return results


class RegressionResults(foe.EvaluationResults):
    """Class that stores the results of a regression evaluation.

    Args:
        ytrue: a list of ground truth values
        ypred: a list of predicted values
        confs (None): an optional list of confidences for the predictions
        eval_key (None): the evaluation key of the evaluation
        gt_field (None): the name of the ground truth field
        pred_field (None): the name of the predictions field
        ids (None): a list of sample or frame IDs corresponding to the
            regressions
        missing (None): a missing value. Any None-valued regressions are
            given this value for results purposes
        samples (None): the :class:`fiftyone.core.collections.SampleCollection`
            for which the results were computed
    """

    def __init__(
        self,
        ytrue,
        ypred,
        confs=None,
        eval_key=None,
        gt_field=None,
        pred_field=None,
        ids=None,
        missing=None,
        samples=None,
    ):
        ytrue, ypred, confs, ids = _parse_values(
            ytrue, ypred, confs, ids, missing=missing
        )

        self.ytrue = ytrue
        self.ypred = ypred
        self.confs = confs
        self.eval_key = eval_key
        self.gt_field = gt_field
        self.pred_field = pred_field
        self.ids = ids
        self.missing = missing

        self._samples = samples

    def metrics(self, weights=None):
        """Computes various popular regression metrics for the results.

        The computed metrics are:

        -   Mean squared error: :func:`sklearn:sklearn.metrics.mean_squared_error`
        -   Root mean squared error: :func:`sklearn:sklearn.metrics.mean_squared_error`
        -   Mean absolute error: :func:`sklearn:sklearn.metrics.mean_absolute_error`
        -   Median absolute error: :func:`sklearn:sklearn.metrics.median_absolute_error`
        -   R^2 score: :func:`sklearn:sklearn.metrics.r2_score`
        -   Explained variance score: :func:`sklearn:sklearn.metrics.explained_variance_score`
        -   Max error: :func:`sklearn:sklearn.metrics.max_error`
        -   Support: the number of examples

        Args:
            weights (None): an optional list of weights for each example

        Returns:
            a dict
        """
        yt = self.ytrue
        yp = self.ypred
        w = weights

        if yt.size > 0:
            mse = skm.mean_squared_error(yt, yp, sample_weight=w)
            rmse = np.sqrt(mse)
            mae = skm.mean_absolute_error(yt, yp, sample_weight=w)
            median_absolute_error = skm.median_absolute_error(yt, yp)
            r2_score = skm.r2_score(yt, yp, sample_weight=w)
            ev_score = skm.explained_variance_score(yt, yp, sample_weight=w)
            max_error = skm.max_error(yt, yp)
            support = len(yt)
        else:
            mse = 0.0
            rmse = 0.0
            mae = 0.0
            median_absolute_error = 0.0
            r2_score = 0.0
            ev_score = 0.0
            max_error = 0.0
            support = 0

        return {
            "mean_squared_error": mse,
            "root_mean_squared_error": rmse,
            "mean_absolute_error": mae,
            "median_absolute_error": median_absolute_error,
            "r2_score": r2_score,
            "explained_variance_score": ev_score,
            "max_error": max_error,
            "support": support,
        }

    def print_metrics(self, weights=None, digits=2):
        """Prints the regression metrics computed via :meth:`metrics`.

        Args:
            weights (None): an optional list of weights for each example
            digits (2): the number of digits of precision to print
        """
        metrics = self.metrics(weights=weights)
        _print_dict_as_table(metrics, digits)

    def plot_results(
        self, labels=None, sizes=None, backend="plotly", **kwargs
    ):
        """Plots the regression results.

        You can use the ``labels`` parameters to define a coloring for the
        points, and you can use the ``sizes`` parameter to scale the sizes of
        the points.

        You can attach plots generated by this method to an App session via its
        :attr:`fiftyone.core.session.Session.plots` attribute, which will
        automatically sync the session's view with the currently selected
        points in the plot.

        Args:
            labels (None): data to use to color the points. Can be any of the
                following:

                -   the name of a sample field or ``embedded.field.name`` of
                    from which to extract numeric or string values
                -   a :class:`fiftyone.core.expressions.ViewExpression`
                    defining numeric or string values to extract via
                    :meth:`fiftyone.core.collections.SampleCollection.values`
                -   a list or array-like of numeric or string values (or lists
                    of lists for frame-level regressions)
            sizes (None): data to use to scale the sizes of the points. Can be
                any of the following:

                -   the name of a sample field or ``embedded.field.name`` from
                    which to extract numeric values
                -   a :class:`fiftyone.core.expressions.ViewExpression`
                    defining numeric values to extract via
                    :meth:`fiftyone.core.collections.SampleCollection.values`
                -   a list or array-like of numeric values (or lists of lists
                    for frame-level regressions)
            backend ("plotly"): the plotting backend to use. Supported values
                are ``("plotly", "matplotlib")``
            **kwargs: keyword arguments for the backend plotting method:

                -   "plotly" backend: :meth:`fiftyone.core.plots.plotly.plot_regressions`
                -   "matplotlib" backend: :meth:`fiftyone.core.plots.matplotlib.plot_regressions`

        Returns:
            an :class:`fiftyone.core.plots.base.InteractivePlot`
        """
        return fop.plot_regressions(
            self.ytrue,
            self.ypred,
            samples=self._samples,
            ids=self.ids,
            labels=labels,
            sizes=sizes,
            gt_field=self.gt_field,
            pred_field=self.pred_field,
            backend=backend,
            **kwargs,
        )

    @classmethod
    def _from_dict(cls, d, samples, config, **kwargs):
        ytrue = d["ytrue"]
        ypred = d["ypred"]
        confs = d.get("confs", None)
        eval_key = d.get("eval_key", None)
        gt_field = d.get("gt_field", None)
        pred_field = d.get("pred_field", None)
        ids = d.get("ids", None)
        missing = d.get("missing", None)
        return cls(
            ytrue,
            ypred,
            confs=confs,
            eval_key=eval_key,
            gt_field=gt_field,
            pred_field=pred_field,
            ids=ids,
            missing=missing,
            samples=samples,
            **kwargs,
        )


def _parse_config(pred_field, gt_field, method, **kwargs):
    if method is None:
        method = "simple"

    if method == "simple":
        return SimpleEvaluationConfig(pred_field, gt_field, **kwargs)

    raise ValueError("Unsupported evaluation method '%s'" % method)


def _safe_mean(values):
    values = [v for v in values if v is not None]
    return np.mean(values) if values else None


def _parse_values(ytrue, ypred, *args, missing=None):
    _ytrue = []
    _ypred = []
    _valid = []
    missing_count = 0

    for yt, yp in zip(ytrue, ypred):
        v = yt is not None and yp is not None

        if missing is None:
            _valid.append(v)

        if v:
            _ytrue.append(yt)
            _ypred.append(yp)
        else:
            missing_count += 1
            if missing is not None:
                if yt is None:
                    yt = missing

                if yp is None:
                    yp = missing

                _ytrue.append(yt)
                _ypred.append(yp)

    found_missing = missing_count > 0

    _ytrue = np.array(_ytrue)
    _ypred = np.array(_ypred)

    if found_missing and missing is None:
        logger.warning(
            "Ignoring %d examples with either missing ground truth or "
            "predictions",
            missing_count,
        )

        valid = np.array(_valid)
        args = [np.asarray(a)[valid] if a is not None else a for a in args]
    else:
        args = [np.asarray(a) if a is not None else a for a in args]

    return (_ytrue, _ypred, *args)


def _print_dict_as_table(d, digits):
    fmt = "%%.%df" % digits
    records = []
    for k, v in d.items():
        k = k.replace("_", " ")
        if isinstance(v, numbers.Integral):
            v = str(v)
        else:
            v = fmt % v

        records.append((k, v))

    print(tabulate(records, tablefmt="plain", numalign="left"))



"""
FiftyOne's migration interface.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import types

from .runner import (
    get_database_revision,
    get_dataset_revision,
    migrate_all,
    migrate_database_if_necessary,
    migrate_dataset_if_necessary,
    needs_migration,
)

# This enables Sphinx refs to directly use paths imported here
__all__ = [
    k
    for k, v in globals().items()
    if not k.startswith("_") and not isinstance(v, types.ModuleType)
]



"""
FiftyOne config.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import logging
import os

try:
    from importlib import metadata as importlib_metadata  # Python 3.8
except ImportError:
    import importlib_metadata  # Python < 3.8

import pytz

import eta
import eta.core.config as etac

import fiftyone.constants as foc
import fiftyone.core.utils as fou

fop = fou.lazy_import("fiftyone.core.plots.plotly")


logger = logging.getLogger(__name__)


class Config(etac.Config):
    """Base class for JSON serializable config classes."""

    def __repr__(self):
        return self.__str__()


class Configurable(etac.Configurable):
    """Base class for classes that can be initialized with a :class:`Config`
    instance that configures their behavior.

    :class:`Configurable` subclasses must obey the following rules:

        (a) Configurable class ``Foo`` has an associated Config class
            ``FooConfig`` that is importable from the same namespace as ``Foo``

        (b) Configurable class ``Foo`` must be initializable via the syntax
            ``Foo(config)``, where config is a ``FooConfig`` instance

    Args:
        config: a :class:`Config`
    """

    def __init__(self, config):
        self.validate(config)
        self.config = config


class EnvConfig(etac.EnvConfig):
    def __repr__(self):
        return self.__str__()


class FiftyOneConfig(EnvConfig):
    """FiftyOne configuration settings."""

    def __init__(self, d=None):
        if d is None:
            d = {}

        self.database_uri = self.parse_string(
            d, "database_uri", env_var="FIFTYONE_DATABASE_URI", default=None
        )
        self.database_validation = self.parse_bool(
            d,
            "database_validation",
            env_var="FIFTYONE_DATABASE_VALIDATION",
            default=True,
        )
        self.database_admin = self.parse_bool(
            d,
            "database_admin",
            env_var="FIFTYONE_DATABASE_ADMIN",
            default=True,
        )
        self.database_dir = self.parse_path(
            d,
            "database_dir",
            env_var="FIFTYONE_DATABASE_DIR",
            default=foc.DEFAULT_DB_DIR,
        )
        self.database_name = self.parse_string(
            d,
            "database_name",
            env_var="FIFTYONE_DATABASE_NAME",
            default="fiftyone",
        )
        self.dataset_zoo_dir = self.parse_path(
            d,
            "dataset_zoo_dir",
            env_var="FIFTYONE_DATASET_ZOO_DIR",
            default=None,
        )
        self.model_zoo_dir = self.parse_path(
            d, "model_zoo_dir", env_var="FIFTYONE_MODEL_ZOO_DIR", default=None
        )
        self.module_path = self.parse_string_array(
            d,
            "module_path",
            env_var="FIFTYONE_MODULE_PATH",
            default=None,
        )
        self.plugins_dir = self.parse_string(
            d, "plugins_dir", env_var="FIFTYONE_PLUGINS_DIR", default=None
        )
        self.dataset_zoo_manifest_paths = self.parse_path_array(
            d,
            "dataset_zoo_manifest_paths",
            env_var="FIFTYONE_DATASET_ZOO_MANIFEST_PATHS",
            default=None,
        )
        self.model_zoo_manifest_paths = self.parse_path_array(
            d,
            "model_zoo_manifest_paths",
            env_var="FIFTYONE_MODEL_ZOO_MANIFEST_PATHS",
            default=None,
        )
        self.default_dataset_dir = self.parse_path(
            d,
            "default_dataset_dir",
            env_var="FIFTYONE_DEFAULT_DATASET_DIR",
            default=None,
        )
        self.default_ml_backend = self.parse_string(
            d,
            "default_ml_backend",
            env_var="FIFTYONE_DEFAULT_ML_BACKEND",
            default=None,
        )
        self.default_batch_size = self.parse_int(
            d,
            "default_batch_size",
            env_var="FIFTYONE_DEFAULT_BATCH_SIZE",
            default=None,
        )
        self.default_sequence_idx = self.parse_string(
            d,
            "default_sequence_idx",
            env_var="FIFTYONE_DEFAULT_SEQUENCE_IDX",
            default="%06d",
        )
        self.default_image_ext = self.parse_string(
            d,
            "default_image_ext",
            env_var="FIFTYONE_DEFAULT_IMAGE_EXT",
            default=".jpg",
        )
        self.default_video_ext = self.parse_string(
            d,
            "default_video_ext",
            env_var="FIFTYONE_DEFAULT_VIDEO_EXT",
            default=".mp4",
        )
        self.default_app_port = self.parse_int(
            d,
            "default_app_port",
            env_var="FIFTYONE_DEFAULT_APP_PORT",
            default=5151,
        )
        self.default_app_address = self.parse_string(
            d,
            "default_app_address",
            env_var="FIFTYONE_DEFAULT_APP_ADDRESS",
            default="localhost",
        )
        self.desktop_app = self.parse_bool(
            d,
            "desktop_app",
            env_var="FIFTYONE_DESKTOP_APP",
            default=False,
        )
        self.logging_level = self.parse_string(
            d,
            "logging_level",
            env_var="FIFTYONE_LOGGING_LEVEL",
            default="INFO",
        )
        self._show_progress_bars = None  # declare
        self.show_progress_bars = self.parse_bool(
            d,
            "show_progress_bars",
            env_var="FIFTYONE_SHOW_PROGRESS_BARS",
            default=True,
        )
        self.do_not_track = self.parse_bool(
            d,
            "do_not_track",
            env_var="FIFTYONE_DO_NOT_TRACK",
            default=False,
        )
        self.requirement_error_level = self.parse_int(
            d,
            "requirement_error_level",
            env_var="FIFTYONE_REQUIREMENT_ERROR_LEVEL",
            default=0,
        )
        self.timezone = self.parse_string(
            d, "timezone", env_var="FIFTYONE_TIMEZONE", default=None
        )

        self._init()

    @property
    def show_progress_bars(self):
        return self._show_progress_bars

    @show_progress_bars.setter
    def show_progress_bars(self, value):
        self._show_progress_bars = value
        try:
            # Keep ETA config in-sync
            eta.config.show_progress_bars = value
        except:
            pass

    def attributes(self):
        # Includes `show_progress_bars`
        return super().custom_attributes(dynamic=True)

    def _init(self):
        if self.default_dataset_dir is None:
            self.default_dataset_dir = os.path.join(
                os.path.expanduser("~"), "fiftyone"
            )

        if self.dataset_zoo_dir is None:
            self.dataset_zoo_dir = self.default_dataset_dir

        if self.model_zoo_dir is None:
            self.model_zoo_dir = os.path.join(
                self.default_dataset_dir, "__models__"
            )

        if self.default_ml_backend is None:
            installed_packages = _get_installed_packages()

            if "torch" in installed_packages:
                self.default_ml_backend = "torch"
            elif "tensorflow" in installed_packages:
                self.default_ml_backend = "tensorflow"

        if self.default_ml_backend is not None:
            self.default_ml_backend = self.default_ml_backend.lower()

        if self.module_path is not None:
            for idx, module_name in enumerate(self.module_path):
                try:
                    __import__(module_name)
                except ImportError as e:
                    logger.warning(
                        "Failed to import fiftyone.config.module_path[%d]: %s",
                        idx,
                        e,
                    )

        if self.timezone and self.timezone.lower() not in {"local", "utc"}:
            try:
                pytz.timezone(self.timezone)
            except:
                logger.warning("Ignoring invalid timezone '%s'", self.timezone)
                self.timezone = None


class FiftyOneConfigError(etac.EnvConfigError):
    """Exception raised when a FiftyOne configuration issue is encountered."""

    pass


class AppConfig(EnvConfig):
    """FiftyOne App configuration settings."""

    def __init__(self, d=None):
        if d is None:
            d = {}

        self.color_by = self.parse_string(
            d,
            "color_by",
            env_var="FIFTYONE_APP_COLOR_BY",
            default="field",
        )
        self.color_pool = self.parse_string_array(
            d,
            "color_pool",
            env_var="FIFTYONE_APP_COLOR_POOL",
            default=foc.DEFAULT_APP_COLOR_POOL,
        )
        self.colorscale = self.parse_string(
            d,
            "colorscale",
            env_var="FIFTYONE_APP_COLORSCALE",
            default="viridis",
        )
        self.grid_zoom = self.parse_int(
            d, "grid_zoom", env_var="FIFTYONE_APP_GRID_ZOOM", default=5
        )
        self.loop_videos = self.parse_bool(
            d,
            "loop_videos",
            env_var="FIFTYONE_APP_LOOP_VIDEOS",
            default=False,
        )
        self.multicolor_keypoints = self.parse_bool(
            d,
            "multicolor_keypoints",
            env_var="FIFTYONE_APP_MULTICOLOR_KEYPOINTS",
            default=False,
        )
        self.notebook_height = self.parse_int(
            d,
            "notebook_height",
            env_var="FIFTYONE_APP_NOTEBOOK_HEIGHT",
            default=800,
        )
        self.show_confidence = self.parse_bool(
            d,
            "show_confidence",
            env_var="FIFTYONE_APP_SHOW_CONFIDENCE",
            default=True,
        )
        self.show_index = self.parse_bool(
            d,
            "show_index",
            env_var="FIFTYONE_APP_SHOW_INDEX",
            default=True,
        )
        self.show_label = self.parse_bool(
            d,
            "show_label",
            env_var="FIFTYONE_APP_SHOW_LABEL",
            default=True,
        )
        self.show_skeletons = self.parse_bool(
            d,
            "show_skeletons",
            env_var="FIFTYONE_APP_SHOW_SKELETONS",
            default=True,
        )
        self.show_tooltip = self.parse_bool(
            d,
            "show_tooltip",
            env_var="FIFTYONE_APP_SHOW_TOOLTIP",
            default=True,
        )
        self.sidebar_mode = self.parse_string(
            d,
            "sidebar_mode",
            env_var="FIFTYONE_APP_SIDEBAR_MODE",
            default="best",
        )
        self.theme = self.parse_string(
            d,
            "theme",
            env_var="FIFTYONE_APP_THEME",
            default="browser",
        )
        self.use_frame_number = self.parse_bool(
            d,
            "use_frame_number",
            env_var="FIFTYONE_APP_USE_FRAME_NUMBER",
            default=False,
        )
        self.plugins = d.get("plugins", {})

        self._init()

    def get_colormap(self, colorscale=None, n=256, hex_strs=False):
        """Generates a continuous colormap with the specified number of colors
        from the given colorscale.

        The provided ``colorscale`` can be any of the following:

        -   The string name of any colorscale recognized by plotly. See
            https://plotly.com/python/colorscales for possible options

        -   A manually-defined colorscale like the following::

                [
                    [0.000, "rgb(165,0,38)"],
                    [0.111, "rgb(215,48,39)"],
                    [0.222, "rgb(244,109,67)"],
                    [0.333, "rgb(253,174,97)"],
                    [0.444, "rgb(254,224,144)"],
                    [0.555, "rgb(224,243,248)"],
                    [0.666, "rgb(171,217,233)"],
                    [0.777, "rgb(116,173,209)"],
                    [0.888, "rgb(69,117,180)"],
                    [1.000, "rgb(49,54,149)"],
                ]

        The colorscale will be sampled evenly at the required resolution in
        order to generate the colormap.

        Args:
            colorscale (None): a valid colorscale. See above for possible
                options. By default, :attr:`colorscale` is used
            n (256): the desired number of colors
            hex_strs (False): whether to return ``#RRGGBB`` hex strings rather
                than ``(R, G, B)`` tuples

        Returns:
            a list of ``(R, G, B)`` tuples in `[0, 255]`, or, if ``hex_strs``
            is True, a list of `#RRGGBB` strings
        """
        if colorscale is None:
            colorscale = self.colorscale

        return fop.get_colormap(colorscale, n=n, hex_strs=hex_strs)

    def _init(self):
        supported_color_bys = {"field", "instance", "label"}
        default_color_by = "field"
        if self.color_by not in supported_color_bys:
            logger.warning(
                "Invalid color_by=%s. Must be one of %s. Defaulting to '%s'",
                self.color_by,
                supported_color_bys,
                default_color_by,
            )
            self.color_by = default_color_by

        supported_sidebar_modes = {"all", "best", "fast"}
        default_sidebar_mode = "best"
        if self.sidebar_mode not in supported_sidebar_modes:
            logger.warning(
                "Invalid sidebar_mode=%s. Must be one of %s. Defaulting to '%s'",
                self.sidebar_mode,
                supported_sidebar_modes,
                default_sidebar_mode,
            )
            self.sidebar_mode = default_sidebar_mode

        supported_themes = {"browser", "dark", "light"}
        default_theme = "browser"
        if self.theme not in supported_themes:
            logger.warning(
                "Invalid theme=%s. Must be one of %s. Defaulting to '%s'",
                self.theme,
                supported_themes,
                default_theme,
            )
            self.theme = default_theme

        if self.grid_zoom < 0 or self.grid_zoom > 10:
            logger.warning(
                "`grid_zoom` must be in [0, 10]; found %d", self.grid_zoom
            )
            self.grid_zoom = 5

        if "MAPBOX_TOKEN" in os.environ:
            try:
                _set_nested_dict_value(
                    self.plugins,
                    "map.mapboxAccessToken",
                    os.environ["MAPBOX_TOKEN"],
                )
            except Exception as e:
                logger.warning("Failed to set mapbox token: %s", e)


class AppConfigError(etac.EnvConfigError):
    """Exception raised when an invalid :class:`AppConfig` instance is
    encountered.
    """

    pass


class AnnotationConfig(EnvConfig):
    """FiftyOne annotation configuration settings."""

    _BUILTIN_BACKENDS = {
        "cvat": {
            "config_cls": "fiftyone.utils.cvat.CVATBackendConfig",
            "url": "https://app.cvat.ai",
        },
        "labelbox": {
            "config_cls": "fiftyone.utils.labelbox.LabelboxBackendConfig",
            "url": "https://labelbox.com",
        },
        "labelstudio": {
            "config_cls": "fiftyone.utils.labelstudio.LabelStudioBackendConfig",
            "url": "https://labelstud.io",
        },
    }

    def __init__(self, d=None):
        if d is None:
            d = {}

        self.default_backend = self.parse_string(
            d,
            "default_backend",
            env_var="FIFTYONE_ANNOTATION_DEFAULT_BACKEND",
            default="cvat",
        )

        self.backends = self._parse_backends(d)

    def _parse_backends(self, d):
        d = d.get("backends", {})
        env_vars = dict(os.environ)

        #
        # `FIFTYONE_ANNOTATION_BACKENDS` can be used to declare which backends
        # are exposed. This may exclude builtin backends and/or declare new
        # backends
        #

        if "FIFTYONE_ANNOTATION_BACKENDS" in env_vars:
            backends = env_vars["FIFTYONE_ANNOTATION_BACKENDS"].split(",")

            # Declare new backends and omit any others not in `backends`
            d = {backend: d.get(backend, {}) for backend in backends}
        else:
            backends = sorted(self._BUILTIN_BACKENDS.keys())

            # Declare builtin backends if necessary
            for backend in backends:
                if backend not in d:
                    d[backend] = {}

        #
        # Extract parameters from any environment variables of the form
        # `FIFTYONE_<BACKEND>_<PARAMETER>`
        #

        for backend, parameters in d.items():
            prefix = "FIFTYONE_%s_" % backend.upper()
            for env_name, env_value in env_vars.items():
                if env_name.startswith(prefix):
                    name = env_name[len(prefix) :].lower()
                    value = _parse_env_value(env_value)
                    parameters[name] = value

        #
        # Set default parameters for builtin annotation backends
        #

        for backend, defaults in self._BUILTIN_BACKENDS.items():
            if backend not in d:
                continue

            d_backend = d[backend]
            for name, value in defaults.items():
                if name not in d_backend:
                    d_backend[name] = value

        return d


def locate_config():
    """Returns the path to the :class:`FiftyOneConfig` on disk.

    The default location is ``~/.fiftyone/config.json``, but you can override
    this path by setting the ``FIFTYONE_CONFIG_PATH`` environment variable.

    Note that a config file may not actually exist on disk in the default
    location, in which case the default config settings will be used.

    Returns:
        the path to the :class:`FiftyOneConfig` on disk

    Raises:
        OSError: if the config path has been customized but the file does not
            exist on disk
    """
    if "FIFTYONE_CONFIG_PATH" not in os.environ:
        return foc.FIFTYONE_CONFIG_PATH

    config_path = os.environ["FIFTYONE_CONFIG_PATH"]
    if not os.path.isfile(config_path):
        raise OSError("Config file '%s' not found" % config_path)

    return config_path


def locate_app_config():
    """Returns the path to the :class:`AppConfig` on disk.

    The default location is ``~/.fiftyone/app_config.json``, but you can
    override this path by setting the ``FIFTYONE_APP_CONFIG_PATH`` environment
    variable.

    Note that a config file may not actually exist on disk in the default
    location, in which case the default config settings will be used.

    Returns:
        the path to the :class:`AppConfig` on disk

    Raises:
        OSError: if the App config path has been customized but the file does
            not exist on disk
    """
    if "FIFTYONE_APP_CONFIG_PATH" not in os.environ:
        return foc.FIFTYONE_APP_CONFIG_PATH

    config_path = os.environ["FIFTYONE_APP_CONFIG_PATH"]
    if not os.path.isfile(config_path):
        raise OSError("App config file '%s' not found" % config_path)

    return config_path


def locate_annotation_config():
    """Returns the path to the :class:`AnnotationConfig` on disk.

    The default location is ``~/.fiftyone/annotation_config.json``, but you can
    override this path by setting the ``FIFTYONE_ANNOTATION_CONFIG_PATH``
    environment variable.

    Note that a config file may not actually exist on disk in the default
    location, in which case the default config settings will be used.

    Returns:
        the path to the :class:`AnnotationConfig` on disk

    Raises:
        OSError: if the annotation config path has been customized but the file
            does not exist on disk
    """
    if "FIFTYONE_ANNOTATION_CONFIG_PATH" not in os.environ:
        return foc.FIFTYONE_ANNOTATION_CONFIG_PATH

    config_path = os.environ["FIFTYONE_ANNOTATION_CONFIG_PATH"]
    if not os.path.isfile(config_path):
        raise OSError("Annotation config file '%s' not found" % config_path)

    return config_path


def load_config():
    """Loads the FiftyOne config.

    Returns:
        a :class:`FiftyOneConfig` instance
    """
    config_path = locate_config()
    if os.path.isfile(config_path):
        return FiftyOneConfig.from_json(config_path)

    return FiftyOneConfig()


def load_app_config():
    """Loads the FiftyOne App config.

    Returns:
        an :class:`AppConfig` instance
    """
    app_config_path = locate_app_config()
    if os.path.isfile(app_config_path):
        return AppConfig.from_json(app_config_path)

    return AppConfig()


def load_annotation_config():
    """Loads the FiftyOne annotation config.

    Returns:
        an :class:`AnnotationConfig` instance
    """
    annotation_config_path = locate_annotation_config()
    if os.path.isfile(annotation_config_path):
        return AnnotationConfig.from_json(annotation_config_path)

    return AnnotationConfig()


def _parse_env_value(value):
    try:
        return int(value)
    except:
        pass

    try:
        return float(value)
    except:
        pass

    if value in ("True", "true"):
        return True

    if value in ("False", "false"):
        return False

    if value in ("None", ""):
        return None

    if "," in value:
        return [_parse_env_value(v) for v in value.split(",")]

    return value


def _get_installed_packages():
    try:
        return set(
            d.metadata["Name"] for d in importlib_metadata.distributions()
        )
    except:
        logger.debug("Failed to get installed packages")
        return set()


def _set_nested_dict_value(d, path, value):
    keys = path.split(".")

    for key in keys[:-1]:
        if key not in d:
            d[key] = {}

        d = d[key]

    d[keys[-1]] = value
